/*
 * ProFTPD: mod_mysql -- Support for auth and up/download ratios with mysql.
 * Time-stamp: <1999-06-14 12:27:47 root>
 * Copyright (c) 1998-1999 Johnie Ingram.
 *  
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307, USA.
 */

/* -- DO NOT MODIFY THE LINE BELOW UNLESS YOU FEEL LIKE IT --
 * $Libraries: -lm /usr/lib/libmysqlclient.a $
 */

/* This is mod_mysql 0.3, contrib software for proftpd 1.2.0pre3 and above.
   For more information contact Johnie Ingram <johnie@netgod.net>.

   History Log:

   * 1999-06-01: v0.3: fixed segfault, changed 'strncmp' typo which
     should have been copy.  Made uid/gid support optional.
   
   * 1999-05-03: v0.2: Removed dead code, fix bug with interaction
     with anon-user support.  Added MySQLLogHits and MySQLHomedir
     directives.  Fixed atoi() invocation that could segfault.
     Copious debugging code added.

   * 1999-04-09: v0.1: Initial attempt.

*/

#include "conf.h"
#include <mysql/mysql.h>
#ifdef HAVE_CRYPT_H
#include <crypt.h>
#endif
#include "mod_mysql.h"

/* *INDENT-OFF* */

/* The Debian defintion of nobody/nogroup, minus 1 (used in getgwnam).  */
#define MOD_MYSQL_MAGIC_USER 65533
#define MOD_MYSQL_MAGIC_GROUP 65533

/* A uid or gid less than this is mapped to the magic numbers above
   instead of simply rejected (which is arguably better, hmm.) */
#define MOD_MYSQL_MIN_ID 999

/* Used to be "mod_mysql" for testing. */
#define MOD_MYSQL_MAGIC_SHELL "/bin/sh"

/* Maximum username field to expect, etc. */
#define ARBITRARY_MAX                   128

static MYSQL mod_mysql_server;

static struct {
  MYSQL* sqldb;

  char user [ARBITRARY_MAX];
  struct passwd pw;           /* Scratch space for the getpwnam call. */
  char *pass;                 /* The password from db. */

  char *homedir;              /* Set when getpwnam is called. */
  int pwnamed;                /* Set when getpwnam is called. */
  int sqled;                  /* Set if sql auth was used. */

  char *sql_host;             /* Data for connecting, set by MySQLInfo.  */
  char *sql_user;
  char *sql_pass;
  char *sql_dbname;

  char *sql_usertable;        /* CREATE TABLE users ( */
  char *sql_userid;           /* userid varchar(50) NOT NULL, */
  char *sql_passwd;           /* passwd varchar(15), */

  char *sql_uid;              /* uid int(5), */
  char *sql_gid;              /* gid int(5), */

  char *sql_fstor;            /* fstor int(11) NOT NULL DEFAULT '0', */
  char *sql_fretr;            /* fretr int(11) NOT NULL DEFAULT '0', */
  char *sql_bstor;            /* bstor int(11) NOT NULL DEFAULT '0', */
  char *sql_bretr;            /* bretr int(11) NOT NULL DEFAULT '0', */

  char *sql_fhdir;            /* fhdir varchar(255), */

  char *sql_fhost;            /* fhost varchar(50), */
  char *sql_faddr;            /* faddr char(15), */
  char *sql_ftime;            /* ftime timestamp, */

  char *sql_fcdir;            /* fcdir varchar(255), */

  char *sql_frate;            /* fstor int(11) NOT NULL DEFAULT '5', */
  char *sql_fcred;            /* fretr int(2) NOT NULL DEFAULT '15', */
  char *sql_brate;            /* bstor int(11) NOT NULL DEFAULT '5', */
  char *sql_bcred;            /* bretr int(2) NOT NULL DEFAULT '150000', */

  char *sql_flogs;            /* flogs int(11) NOT NULL DEFAULT '0', */

  char *sql_hittable;         /* FIXME: undocumented */
  char *sql_dir;
  char *sql_filename;
  char *sql_hits;

  char *loginmsg;

} g;

/* *INDENT-ON* */


/* **************************************************************** */

int modmysql_disconnect(void)
{
  if (g.sqldb)
  {
    log_debug(DEBUG5, "mysql: disconnecting: %s",
	mysql_stat(g.sqldb));
    mysql_close(g.sqldb);
  }
  g.sqldb = NULL;
  return 0;
}

int modmysql_connect(void)
{
  modmysql_disconnect();
  if (!(g.sqldb = mysql_connect(&mod_mysql_server, g.sql_host,
	      g.sql_user, g.sql_pass)))
  {
    log_pri(LOG_ERR, "mysql: client %s failed connect to %s@%s",
	mysql_get_client_info(), g.sql_user, g.sql_host);
    return -1;
  }
  log_debug(DEBUG5, "mysql: connect OK %s -> %s (%s@%s)",
      mysql_get_client_info(), mysql_get_server_info(g.sqldb),
      g.sql_user, g.sql_host);
  return 0;
}

int modmysql_query(cmd_rec * cmd, const char *query)
{
  int error = 1;

  if (!g.sql_host)
    return -1;

  block_signals();

  if (g.sqldb)
    mysql_select_db(g.sqldb, g.sql_dbname);
  if (!g.sqldb
      || ((error = mysql_query(g.sqldb, query))
	  && !strcasecmp(mysql_error(g.sqldb), "mysql server has gone")))
  {
    /* We need to restart the server link. */
    if (g.sqldb)
      log_pri(LOG_ERR, "mysql: server has wandered off (db %s)",
	  g.sql_dbname);
    modmysql_connect();
    if (!g.sqldb)
      return -1;
    error = mysql_select_db(g.sqldb, g.sql_dbname)
	|| mysql_query(g.sqldb, query);
  }

  unblock_signals();

  if (error)
  {
    log_debug(DEBUG4, "mysql: query failed: \"%s\": %s",
	query, mysql_error(g.sqldb));
    return -1;
  }

  log_debug(DEBUG5, "mysql: query OK: [%s] \"%s\"", g.sql_dbname, query);
  return 0;
}

int modmysql_queryuser(cmd_rec * cmd, const char *user, const char *query,
    int update)
{
  char userbuf[ARBITRARY_MAX];
  char *realquery;

  mysql_escape_string(userbuf, user, strlen(user));
  userbuf[ARBITRARY_MAX - 1] = 0;

  if (update)
    realquery = pstrcat(cmd->tmp_pool, "update ", g.sql_usertable,
	" set ", query, " where ", g.sql_userid, " = '",
	user, "'", NULL);
  else
    realquery = pstrcat(cmd->tmp_pool, "select ", query, " from ",
	g.sql_usertable, " where ", g.sql_userid,
	" = '", user, "'", NULL);
  return modmysql_query(cmd, realquery);
}


/* Returns 1 on successful match, 0 unsuccessful match, -1 on error. */

int modmysql_checkpass(cmd_rec * cmd, const char *user, const char *pass)
{
  int success = 0;
  char *query;
  MYSQL_RES *result;
  MYSQL_ROW sql_row = 0;
  int emptyok, cplain, cmysql, ccrypt;

  emptyok = get_param_int(TOPLEVEL_CONF, "MySQLEmptyPasswords", FALSE);
  cplain = get_param_int(TOPLEVEL_CONF, "MySQLPlaintextPasswords", FALSE);
  ccrypt = get_param_int(TOPLEVEL_CONF, "MySQLEncryptedPasswords", FALSE);
  cmysql = get_param_int(TOPLEVEL_CONF, "MySQLScrambledPasswords", FALSE);

  if (modmysql_queryuser(cmd, user, g.sql_passwd, FALSE))
    return -1;

  if ((result = mysql_store_result(g.sqldb)))
  {
    switch (mysql_num_rows(result))
    {
      case 0:
	return 0;

      case 1:
	sql_row = mysql_fetch_row(result);

	/* Ensure we have a row, and non NULL value. */
	if (!sql_row)
	  break;
	if (!sql_row[0])
	{
	  log_debug(DEBUG4, "mysql: %s auth declined for NULL pass", user);
	  break;
	}

	if (emptyok == TRUE && !strlen(sql_row[0]))
	{
	  log_debug(DEBUG4, "mysql: warning: %s has empty password",
	      user);
	  success = 1;
	}

	if (!success && cplain == TRUE
	    && !strncasecmp(sql_row[0], pass, 10))
	{
	  success = 1;
	}

	if (!success && cmysql == TRUE)
	{
	  char scrambled_pass[32];

	  make_scrambled_password(scrambled_pass, pass);
	  if (!strcmp(scrambled_pass, pass))
	    success = 1;
	}

	/* Deliberate: ccrypt same if TRUE or -1 (unspecified). */
	if (!success && ccrypt)
	{
	  if (!strcmp(query = crypt(pass, sql_row[0]), sql_row[0]))
	    success = 1;
	}

      default:
	break;
    }

    mysql_free_result(result);
  }

  if (!success)
    log_debug(DEBUG5, "mysql: %s auth failed: '%s' != '%s'",
	      user, pass, sql_row ? sql_row[0] : "");
  return success;
}

int modmysql_getstats(cmd_rec * cmd, modmysql_luser * stats)
{
  char buf[ARBITRARY_MAX];
  MYSQL_RES *result;
  MYSQL_ROW sql_row;

  snprintf(buf, sizeof(buf), "%s, %s, %s, %s",
      g.sql_fstor, g.sql_fretr, g.sql_bstor, g.sql_bretr);
  if (modmysql_queryuser(cmd, g.user, buf, FALSE))
    return -1;

  if ((result = mysql_store_result(g.sqldb)))
  {
    memset(stats, 0, sizeof(modmysql_luser));
    switch (mysql_num_rows(result))
    {
      case 1:
	if (!(sql_row = mysql_fetch_row(result)))
	  break;
	if (sql_row[0])
	  stats->fstor = atoi(sql_row[0]);
	if (sql_row[1])
	  stats->fretr = atoi(sql_row[1]);
	if (sql_row[2])
	  stats->bstor = atoi(sql_row[2]);
	if (sql_row[3])
	  stats->bretr = atoi(sql_row[3]);
	break;

      default:
	log_debug(DEBUG4, "mysql: warning: getstats on %s not unique",
	    g.user);
      case 0:
    }
    mysql_free_result(result);
  }

  if (!g.sql_frate)
    return 0;

  snprintf(buf, sizeof(buf), "%s, %s, %s, %s",
      g.sql_frate, g.sql_fcred, g.sql_brate, g.sql_bcred);
  if (modmysql_queryuser(cmd, g.user, buf, FALSE))
    return 0;

  if ((result = mysql_store_result(g.sqldb)))
  {
    switch (mysql_num_rows(result))
    {
      case 1:
	if (!(sql_row = mysql_fetch_row(result)))
	  break;
	if (sql_row[0])
	  stats->frate = atoi(sql_row[0]);
	if (sql_row[1])
	  stats->fcred = atoi(sql_row[1]);
	if (sql_row[2])
	  stats->brate = atoi(sql_row[2]);
	if (sql_row[3])
	  stats->bcred = atoi(sql_row[3]);
	break;

      default:
	log_debug(DEBUG4, "mysql: warning: getratios on %s not unique",
	    g.user);
      case 0:
    }
    mysql_free_result(result);
  }

  return 0;
}

int modmysql_setstats(cmd_rec * cmd, int fstor, int fretr, int bstor, int bretr)
{
  char query[128];
  modmysql_luser stats;

  if (g.sqldb)
  {
    snprintf(query, sizeof(query),
	     "%s = %s + %i, %s = %s + %i, %s = %s + %i, %s = %s + %i",
	     g.sql_fstor, g.sql_fstor, stats.fstor += fstor,
	     g.sql_fretr, g.sql_fretr, stats.fretr += fretr,
	     g.sql_bstor, g.sql_bstor, stats.bstor += bstor,
	     g.sql_bretr, g.sql_bretr, stats.bretr += bretr);
    modmysql_queryuser(cmd, g.user, query, TRUE);
  }
  return 0;
}

const char *modmysql_getuservar(cmd_rec * cmd, const char *user, const char *var)
{
  MYSQL_RES *result;
  MYSQL_ROW sql_row;
  char *success = NULL;

  if (modmysql_queryuser(cmd, user, var, FALSE))
    return success;
  if ((result = mysql_store_result(g.sqldb)))
  {
    if (mysql_num_rows(result) == 1)
    {
      sql_row = mysql_fetch_row(result);
      if (sql_row && sql_row[0])
	success = pstrdup(cmd->tmp_pool, sql_row[0]);
    }
    mysql_free_result(result);
  }
  return success;
}

/* **************************************************************** */

/* Note: This function is called once by the master proftpd process
   before it forks, so thankfully the homedir field is only set in the
   child_init(). */

MODRET auth_cmd_getpwnam(cmd_rec * cmd)
{
  const char *homedir;

  if (!g.sql_fhdir && !g.homedir)
    return DECLINED(cmd);

  /* Only try this for the USER command (is also called twice after
     PASS for some reason. */
  if (!g.homedir && !g.pwnamed++)
  {
    if ((homedir = modmysql_getuservar(cmd, cmd->argv[0], g.sql_fhdir)))
      g.homedir = pstrdup(session.pool, homedir);
  }
  if (!g.homedir)
    return DECLINED(cmd);

  if (!g.pw.pw_name)
  {
    g.pw.pw_name = cmd->argv[0];
    if (g.sql_uid)
      g.pw.pw_uid =
	atoi(modmysql_getuservar(cmd, cmd->argv[0], g.sql_uid) ?: "0");
    if (g.pw.pw_uid < MOD_MYSQL_MIN_ID)
      g.pw.pw_uid = MOD_MYSQL_MAGIC_USER;
    if (g.sql_gid)
      g.pw.pw_gid =
	atoi(modmysql_getuservar(cmd, cmd->argv[0], g.sql_gid) ?: "0");
    if (g.pw.pw_gid < MOD_MYSQL_MIN_ID)
      g.pw.pw_gid = MOD_MYSQL_MAGIC_GROUP;
    g.pw.pw_shell = MOD_MYSQL_MAGIC_SHELL;
    g.pw.pw_dir = (char *)g.homedir;
    log_debug(DEBUG3, "mysql: user \"%s\" (%i/%i) for %s",
	cmd->argv[0], g.pw.pw_uid, g.pw.pw_gid, g.pw.pw_dir);

    /* Copy username so proftpd anon handling won't confuse the issue. */
    strncpy(g.user, g.pw.pw_name, ARBITRARY_MAX - 1);
    g.user[ARBITRARY_MAX] = 0;
  }

  return mod_create_data(cmd, &g.pw);
}

MODRET auth_cmd_auth(cmd_rec * cmd)
{
  char *user, *pass;
  int return_type;
  int retval = AUTH_NOPWD;

  if (!g.sql_passwd)
    return DECLINED(cmd);

  /* Figure out our default return style: Whether or not MySQL should
   * allow other auth modules a shot at this user or not is controlled
   * by the parameter "MySQLAuthoritative".  Like mod_pam this
   * defaults to no.  */
  return_type = get_param_int(TOPLEVEL_CONF,
      "MySQLAuthoritative", FALSE);

  /* Just in case... */
  if (cmd->argc != 2)
    return return_type ? ERROR(cmd) : DECLINED(cmd);
  if ((user = cmd->argv[0]) == NULL)
    return return_type ? ERROR(cmd) : DECLINED(cmd);
  if ((pass = cmd->argv[1]) == NULL)
    return return_type ? ERROR(cmd) : DECLINED(cmd);

  if (!g.pass
      && get_param_int(TOPLEVEL_CONF, "MySQLEmptyPasswords", FALSE) == 2)
    {
      char *query = pstrcat(cmd->tmp_pool, "update ", g.sql_usertable,
		      " set ", g.sql_passwd, " = '", cmd->argv[1],
		      "' where ", g.sql_userid, " = '", g.user, "'", 0);
      log_debug(DEBUG3, "mysql: %s NULL pass set to '%s'",
		user, cmd->argv[1], query);
      modmysql_query(cmd, query);
    }

  if (modmysql_checkpass(cmd, user, pass) != 1)
    return return_type ? ERROR_INT(cmd, retval) : DECLINED(cmd);

  g.sqled++;
  return HANDLED(cmd);
}

static authtable mysql_authtab[] =
{
    {0, "getpwnam", auth_cmd_getpwnam},
    {0, "auth", auth_cmd_auth},

    {0, NULL, NULL}
};

/* **************************************************************** */

MODRET pre_cmd_quit(cmd_rec * cmd)
{
  modmysql_disconnect();
  return DECLINED(cmd);
}

MODRET cmd_user(cmd_rec * cmd)
{
  if (!g.user[0])
    strncpy(g.user, cmd->argv[1], ARBITRARY_MAX - 1);

  if (g.sql_passwd)
    {
      g.pass = (char *) modmysql_getuservar(cmd, cmd->argv[1], g.sql_passwd);
      if (!g.pass
	  && get_param_int(TOPLEVEL_CONF, "MySQLEmptyPasswords", FALSE) == 2)
	if (modmysql_getuservar(cmd, cmd->argv[1], "1"))
	  add_response (R_331, "Changing password for %s -- "
			"this password will be saved.", cmd->argv[1]);
    }
  return DECLINED(cmd);
}

MODRET cmd_pass(cmd_rec * cmd)
{
  if (g.sql_passwd && !g.pass
      && get_param_int(TOPLEVEL_CONF, "MySQLEmptyPasswords", FALSE) == 2)
    add_response(R_230, "\"%s\" is the new user \"%s\" password.",
		 cmd->argv[1], g.user);

  /* This is called before the disconnect() in log_cmd_pass. */
  if (g.sql_fcdir)
    {
      const char *d = modmysql_getuservar(cmd, g.user, g.sql_fcdir);
      if (d)
	add_response(R_230, "\"%s\" was last directory.", d);
    }

  return DECLINED(cmd);
}

MODRET post_cmd_pass(cmd_rec * cmd)
{
  if (g.sql_passwd && g.sqled)
    session.anon_user = session.user = (char *) g.user;
  return DECLINED(cmd);
}

MODRET post_cmd_stor(cmd_rec * cmd)
{
  if (g.sql_fstor)
    modmysql_setstats(cmd, 1, 0, session.xfer.total_bytes, 0);
  return DECLINED(cmd);
}

/* FIXME: make sql_dir optional in config */
MODRET cmd_retr(cmd_rec * cmd)
{
  int i;
  char *path, *filename, *query;
  if (g.sql_hittable)
  {
    path = dir_realpath (cmd->tmp_pool, cmd->argv[1]);
    if (g.sql_dir)
      {
	for (i = strlen (path), filename = path + i;
	     *filename != '/' && i > 1; i--)
	  filename--;
	*filename++ = 0; 
	query = pstrcat(cmd->tmp_pool, "update ", g.sql_hittable,
			" set ", g.sql_hits, " = ", g.sql_hits,
			" + 1 where ", g.sql_dir, " = '", ++path,
			"' and ", g.sql_filename, " = '", filename,
			"'", 0);
      }
    else
      query = pstrcat(cmd->tmp_pool, "update ", g.sql_hittable,
		      " set ", g.sql_hits, " = ", g.sql_hits,
		      " + 1 where ", g.sql_filename, " = '", path,
		      "'", 0);
    modmysql_query(cmd, query);
  }
  return DECLINED(cmd);
}

MODRET post_cmd_retr(cmd_rec * cmd)
{
  if (g.sql_fretr)
    modmysql_setstats(cmd, 0, 1, 0, session.xfer.total_bytes);
  return DECLINED(cmd);
}

MODRET log_cmd_pass(cmd_rec * cmd)
{
  char *query;

  if (g.sql_fhost)
  {
    query = pstrcat(cmd->tmp_pool, g.sql_fhost, " = '",
	session.c->remote_name, "', ", g.sql_faddr,
	" = '", inet_ntoa(*session.c->remote_ipaddr),
	"', ", g.sql_ftime, " = now()", NULL);
    modmysql_queryuser(cmd, g.user, query, TRUE);
  }
  if (g.sql_flogs)
  {
    query = pstrcat(cmd->tmp_pool, g.sql_flogs, " = ", g.sql_flogs,
	" + 1", NULL);
    modmysql_queryuser(cmd, g.user, query, TRUE);
  }

  /* Autononpersistence: disconnect now if no other feature is being used. */
  if (!g.sql_fstor && !g.sql_fcdir && !g.sql_hittable)
    modmysql_disconnect();
  return DECLINED(cmd);
}

MODRET log_cmd_cwd(cmd_rec * cmd)
{
  if (g.sql_fcdir)
  {
    char *query = pstrcat(cmd->tmp_pool, g.sql_fcdir, " = '",
	session.cwd, "'", NULL);
    modmysql_queryuser(cmd, g.user, query, TRUE);
  }
  return DECLINED(cmd);
}

cmdtable mysql_cmdtab[] =
{
/* *INDENT-OFF* */

  { PRE_CMD,  C_QUIT,	G_NONE, pre_cmd_quit, 	FALSE, FALSE },
  { CMD,      C_USER,	G_NONE, cmd_user, 	FALSE, FALSE },
  { CMD,      C_PASS,	G_NONE, cmd_pass, 	FALSE, FALSE },
  { POST_CMD, C_PASS,	G_NONE, post_cmd_pass, 	FALSE, FALSE },
  { POST_CMD, C_STOR,	G_NONE, post_cmd_stor, 	FALSE, FALSE },
  { POST_CMD, C_RETR,	G_NONE, post_cmd_retr, 	FALSE, FALSE },
  { CMD,      C_RETR,	G_NONE, cmd_retr, 	FALSE, FALSE },
  { LOG_CMD,  C_PASS,	G_NONE, log_cmd_pass, 	FALSE, FALSE },
  { LOG_CMD,  C_CWD,	G_NONE, log_cmd_cwd, 	FALSE, FALSE },

  { 0,	      NULL }

/* *INDENT-ON* */
};

/* **************************************************************** */

MODRET set_sqlinfo(cmd_rec * cmd)
{
  CHECK_ARGS(cmd, 4);
  CHECK_CONF(cmd, CONF_ROOT | CONF_GLOBAL);
  add_config_param_str("MySQLInfo", 4,
      (void *)cmd->argv[1], (void *)cmd->argv[2],
      (void *)cmd->argv[3], (void *)cmd->argv[4]);
  return HANDLED(cmd);
}

MODRET set_sqlloghosts(cmd_rec * cmd)
{
  int b;

  CHECK_CONF(cmd, CONF_ROOT | CONF_GLOBAL);
  switch (cmd->argc - 1)
  {
    default:
      CONF_ERROR(cmd, "requires a boolean or 3 field names: "
	  "fhost faddr ftime");
    case 1:
      if ((b = get_boolean(cmd, 1)) == -1)
	CONF_ERROR(cmd, "requires a boolean or 3 field names: "
	    "fhost faddr ftime");
      if (b)
	add_config_param_str("MySQLLogHosts", 3, "fhost", "faddr", "ftime");
      break;

    case 3:
      add_config_param_str("MySQLLogHosts", 3,
	  (void *)cmd->argv[1], (void *)cmd->argv[2],
	  (void *)cmd->argv[3]);
  }
  return HANDLED(cmd);
}

MODRET set_sqllogstats(cmd_rec * cmd)
{
  int b;

  CHECK_CONF(cmd, CONF_ROOT | CONF_GLOBAL);
  switch (cmd->argc - 1)
  {
    default:
      CONF_ERROR(cmd, "requires a boolean or 4 field names: "
	  "fstor fretr bstor bretr");
    case 1:
      if ((b = get_boolean(cmd, 1)) == -1)
	CONF_ERROR(cmd, "requires a boolean or 4 field names: "
	    "fstor fretr bstor bretr");
      if (b)
	add_config_param_str("MySQLLogStats", 4,
	    "fstor", "fretr", "bstor", "bretr");
      break;

    case 4:
      add_config_param_str("MySQLLogStats", 4,
	  (void *)cmd->argv[1], (void *)cmd->argv[2],
	  (void *)cmd->argv[3], (void *)cmd->argv[4]);
  }
  return HANDLED(cmd);
}

MODRET set_sqlloghits(cmd_rec * cmd)
{
  CHECK_CONF(cmd, CONF_ROOT | CONF_GLOBAL);
  switch (cmd->argc - 1)
  {
    default:
      CONF_ERROR(cmd, "requires a table or table plus 3 fields: "
	  "[table] dir filename hits");
    case 1:
      add_config_param_str("MySQLLogHits", 4, (void *)cmd->argv[1],
	  "dir", "filename", "hits");
      break;

    case 4:
      add_config_param_str("MySQLLogHits", 4,
	  (void *)cmd->argv[1], (void *)cmd->argv[2],
	  (void *)cmd->argv[3], (void *)cmd->argv[4]);
  }
  return HANDLED(cmd);
}

MODRET set_sqllogdirs(cmd_rec * cmd)
{
  int b;

  CHECK_ARGS(cmd, 1);
  CHECK_CONF(cmd, CONF_ROOT | CONF_GLOBAL);
  if ((b = get_boolean(cmd, 1)) == -1)
    add_config_param_str("MySQLLogDirs", 1, (void *)cmd->argv[1]);
  else if (b)
    add_config_param_str("MySQLLogDirs", 1, "fcdir");
  return HANDLED(cmd);
}

MODRET set_sqlratios(cmd_rec * cmd)
{
  int b;

  CHECK_CONF(cmd, CONF_ROOT | CONF_GLOBAL);
  switch (cmd->argc - 1)
  {
    default:
      CONF_ERROR(cmd, "requires a boolean or 4 field names: "
	  "frate fcred brate bcred");
    case 1:
      if ((b = get_boolean(cmd, 1)) == -1)
	CONF_ERROR(cmd, "requires a boolean or 4 field names: "
	    "frate fcred brate bcred");
      if (b)
	add_config_param_str("MySQLRatios", 4,
	    "frate", "fcred", "brate", "bcred");
      break;

    case 4:
      add_config_param_str("MySQLRatios", 4,
	  (void *)cmd->argv[1], (void *)cmd->argv[2],
	  (void *)cmd->argv[3], (void *)cmd->argv[4]);
  }
  return HANDLED(cmd);
}

MODRET set_sqlempty(cmd_rec * cmd)
{
  int b;
  config_rec *c;

  CHECK_ARGS(cmd, 1);
  CHECK_CONF(cmd, CONF_ROOT | CONF_GLOBAL);

  if ((b = get_boolean(cmd, 1)) == -1)
    {
      if (!strcasecmp (cmd->argv[1], "set"))
	b = 2;
      else
	CONF_ERROR(cmd, "requires 'set' or a boolean value");
    }
  c = add_config_param(cmd->argv[0], 1, (void *)b);
  c->flags |= CF_MERGEDOWN;
  return HANDLED(cmd);
}

MODRET add_globalstr(cmd_rec * cmd)
{
  CHECK_ARGS(cmd, 1);
  CHECK_CONF(cmd, CONF_ROOT | CONF_GLOBAL);
  add_config_param_str(cmd->argv[0], 1, (void *)cmd->argv[1]);
  return HANDLED(cmd);
}

MODRET add_globalbool(cmd_rec * cmd)
{
  int b;
  config_rec *c;

  CHECK_ARGS(cmd, 1);
  CHECK_CONF(cmd, CONF_ROOT | CONF_GLOBAL);

  b = get_boolean(cmd, 1);
  if (b == -1)
    CONF_ERROR(cmd, "requires a boolean value");
  c = add_config_param(cmd->argv[0], 1, (void *)b);
  c->flags |= CF_MERGEDOWN;
  return HANDLED(cmd);
}


static conftable mysql_conftab[] =
{
/* *INDENT-OFF* */

  { "MySQLInfo",                set_sqlinfo,       NULL },
  { "MySQLLogHosts",            set_sqlloghosts,   NULL },
  { "MySQLLogStats",            set_sqllogstats,   NULL },
  { "MySQLLogHits",             set_sqlloghits,    NULL },
  { "MySQLLogDirs",             set_sqllogdirs,    NULL },
  { "MySQLRatios",              set_sqlratios,     NULL },
  { "MySQLEmptyPasswords",      set_sqlempty,      NULL },

  { "MySQLUserTable",           add_globalstr,     NULL },
  { "MySQLUsernameField",       add_globalstr,     NULL },
  { "MySQLUidField",            add_globalstr,     NULL },
  { "MySQLGidField",            add_globalstr,     NULL },
  { "MySQLPasswordField",       add_globalstr,     NULL },
  { "MySQLHomedirField",        add_globalstr,     NULL },
  { "MySQLLoginCountField",     add_globalstr,     NULL },
  { "MySQLHomedir",             add_globalstr,     NULL },

  { "MySQLAuthoritative",       add_globalbool,    NULL },
  { "MySQLEncryptedPasswords",  add_globalbool,    NULL },
  { "MySQLScrambledPasswords",  add_globalbool,    NULL },
  { "MySQLPlaintextPasswords",  add_globalbool,    NULL },

  { NULL, NULL, NULL }

/* *INDENT-ON* */
};

/* **************************************************************** */

static int mysql_child_init()
{
  config_rec *c;

  memset(&g, 0, sizeof(g));
  if (!(c = find_config(CURRENT_CONF, CONF_PARAM, "MySQLInfo", FALSE)))
    return 0;

  g.sql_host = pstrdup(session.pool, c->argv[0]);
  g.sql_user = pstrdup(session.pool, c->argv[1]);
  g.sql_pass = pstrdup(session.pool, c->argv[2]);
  g.sql_dbname = pstrdup(session.pool, c->argv[3]);

  g.sql_passwd = get_param_ptr(TOPLEVEL_CONF, "MySQLPasswordField", FALSE);
  g.sql_flogs = get_param_ptr(TOPLEVEL_CONF, "MySQLLoginCountField", FALSE);
  g.sql_uid = get_param_ptr(TOPLEVEL_CONF, "MySQLUidField", FALSE);
  g.sql_gid = get_param_ptr(TOPLEVEL_CONF, "MySQLGidField", FALSE);

  if (! (g.homedir = get_param_ptr(TOPLEVEL_CONF, "MySQLHomedir", FALSE)))
    {
      g.sql_fhdir = get_param_ptr(TOPLEVEL_CONF, "MySQLHomedirField",
				  FALSE);
    }
  if ((c = find_config(CURRENT_CONF, CONF_PARAM, "MySQLLogHosts", FALSE)))
  {
    g.sql_fhost = c->argv[0];
    g.sql_faddr = c->argv[1];
    g.sql_ftime = c->argv[2];
  }
  if ((c = find_config(CURRENT_CONF, CONF_PARAM, "MySQLLogStats", FALSE)))
  {
    g.sql_fstor = c->argv[0];
    g.sql_fretr = c->argv[1];
    g.sql_bstor = c->argv[2];
    g.sql_bretr = c->argv[3];
  }
  if ((c = find_config(CURRENT_CONF, CONF_PARAM, "MySQLRatios", FALSE)))
  {
    if (!g.sql_fstor)
      log_pri(LOG_WARNING, "mysql: warning: MySQLRatios directive "
	  "ineffective without MySQLLogStats on");
    g.sql_frate = c->argv[0];
    g.sql_fcred = c->argv[1];
    g.sql_brate = c->argv[2];
    g.sql_bcred = c->argv[3];
  }
  if ((c = find_config(CURRENT_CONF, CONF_PARAM, "MySQLLogHits", FALSE)))
  {
    g.sql_hittable = c->argv[0];
    g.sql_dir = c->argv[1];
    g.sql_filename = c->argv[2];
    g.sql_hits = c->argv[3];
  }
  g.sql_fcdir = get_param_ptr(TOPLEVEL_CONF, "MySQLLogDirs", FALSE);

  g.sql_usertable = get_param_ptr(TOPLEVEL_CONF, "MySQLUserTable",
      FALSE) ? : "users";
  g.sql_userid = get_param_ptr(TOPLEVEL_CONF, "MySQLUsernameField",
      FALSE) ? : "userid";

  if (!g.sql_passwd && !g.sql_fhost
      && !g.sql_fstor && !g.sql_fcdir)
    return 0;

  modmysql_connect();
  return 0;
}

module mysql_module =
{
    NULL, NULL,			/* Always NULL */
    0x20,			/* API Version 2.0 */
    "mysql",
    mysql_conftab,		/* SQL configuration handler table */
    mysql_cmdtab,		/* SQL command handler table */
    mysql_authtab,		/* SQL authentication handler table */
    NULL,			/* No initial initialization needed */
    mysql_child_init		/* Post-fork "child mode" init */
};
